@@ -55,7 +55,7 @@ static const char *config_keys[] =
 static int config_keys_num = STATIC_ARRAY_SIZE (config_keys);
 
 static size_t apache_curl_callback (void *buf, size_t size, size_t nmemb,
-		void *stream)
+		void __attribute__((unused)) *stream)
 {
 	size_t len = size * nmemb;
 
@@ -146,16 +146,15 @@ static int init (void)
 	{
 		int status;
 
-		status = snprintf (credentials, sizeof (credentials), "%s:%s",
+		status = ssnprintf (credentials, sizeof (credentials), "%s:%s",
 				user, (pass == NULL) ? "" : pass);
-		if (status >= sizeof (credentials))
+		if ((status < 0) || ((size_t) status >= sizeof (credentials)))
 		{
 			ERROR ("apache plugin: init: Returning an error "
 					"because the credentials have been "
 					"truncated.");
 			return (-1);
 		}
-		credentials[sizeof (credentials) - 1] = '\0';
 
 		curl_easy_setopt (curl, CURLOPT_USERPWD, credentials);
 	}
@@ -198,19 +197,16 @@ static void submit_counter (const char *type, const char *type_instance,
 
 	vl.values = values;
 	vl.values_len = 1;
-	vl.time = time (NULL);
 	sstrncpy (vl.host, hostname_g, sizeof (vl.host));
 	sstrncpy (vl.plugin, "apache", sizeof (vl.plugin));
 	sstrncpy (vl.plugin_instance, "", sizeof (vl.plugin_instance));
+	sstrncpy (vl.type, type, sizeof (vl.type));
 
 	if (type_instance != NULL)
-	{
-		strncpy (vl.type_instance, type_instance,
+		sstrncpy (vl.type_instance, type_instance,
 				sizeof (vl.type_instance));
-		vl.type_instance[sizeof (vl.type_instance) - 1] = '\0';
-	}
 
-	plugin_dispatch_values (type, &vl);
+	plugin_dispatch_values (&vl);
 } /* void submit_counter */
 
 static void submit_gauge (const char *type, const char *type_instance,
@@ -223,19 +219,16 @@ static void submit_gauge (const char *type, const char *type_instance,
 
 	vl.values = values;
 	vl.values_len = 1;
-	vl.time = time (NULL);
 	sstrncpy (vl.host, hostname_g, sizeof (vl.host));
 	sstrncpy (vl.plugin, "apache", sizeof (vl.plugin));
 	sstrncpy (vl.plugin_instance, "", sizeof (vl.plugin_instance));
+	sstrncpy (vl.type, type, sizeof (vl.type));
 
 	if (type_instance != NULL)
-	{
-		strncpy (vl.type_instance, type_instance,
+		sstrncpy (vl.type_instance, type_instance,
 				sizeof (vl.type_instance));
-		vl.type_instance[sizeof (vl.type_instance) - 1] = '\0';
-	}
 
-	plugin_dispatch_values (type, &vl);
+	plugin_dispatch_values (&vl);
 } /* void submit_counter */
 
 static void submit_scoreboard (char *buf)
