@@ -18,6 +18,8 @@
  * Authors:
  *   Florian octo Forster <octo at verplant.org>
  *   Niki W. Waibel <niki.waibel@gmx.net>
+ *   Sebastian Harl <sh at tokkee.org>
+ *   Michał Mirosław <mirq-linux at rere.qmqm.pl>
 **/
 
 #if HAVE_CONFIG_H
@@ -61,18 +63,37 @@ char *sstrncpy (char *dest, const char *src, size_t n)
 	return (dest);
 } /* char *sstrncpy */
 
+int ssnprintf (char *dest, size_t n, const char *format, ...)
+{
+	int ret = 0;
+	va_list ap;
+
+	va_start (ap, format);
+	ret = vsnprintf (dest, n, format, ap);
+	dest[n - 1] = '\0';
+	va_end (ap);
+
+	return (ret);
+} /* int ssnprintf */
+
 char *sstrdup (const char *s)
 {
 	char *r;
+	size_t sz;
 
 	if (s == NULL)
 		return (NULL);
 
-	if((r = strdup (s)) == NULL)
+	/* Do not use `strdup' here, because it's not specified in POSIX. It's
+	 * ``only'' an XSI extension. */
+	sz = strlen (s) + 1;
+	r = (char *) malloc (sizeof (char) * sz);
+	if (r == NULL)
 	{
-		ERROR ("Not enough memory.");
+		ERROR ("sstrdup: Out of memory.");
 		exit (3);
 	}
+	memcpy (r, s, sizeof (char) * sz);
 
 	return (r);
 } /* char *sstrdup */
@@ -91,7 +112,7 @@ char *sstrerror (int errnum, char *buf, size_t buflen)
 		pthread_mutex_lock (&strerror_r_lock);
 
 		temp = strerror (errnum);
-		strncpy (buf, temp, buflen);
+		sstrncpy (buf, temp, buflen);
 
 		pthread_mutex_unlock (&strerror_r_lock);
 	}
@@ -104,9 +125,9 @@ char *sstrerror (int errnum, char *buf, size_t buflen)
 		if (buf[0] == '\0')
 		{
 			if ((temp != NULL) && (temp != buf) && (temp[0] != '\0'))
-				strncpy (buf, temp, buflen);
+				sstrncpy (buf, temp, buflen);
 			else
-				strncpy (buf, "strerror_r did not return "
+				sstrncpy (buf, "strerror_r did not return "
 						"an error message", buflen);
 		}
 	}
@@ -115,13 +136,12 @@ char *sstrerror (int errnum, char *buf, size_t buflen)
 #else
 	if (strerror_r (errnum, buf, buflen) != 0)
 	{
-		snprintf (buf, buflen, "Error #%i; "
+		ssnprintf (buf, buflen, "Error #%i; "
 				"Additionally, strerror_r failed.",
 				errnum);
 	}
 #endif /* STRERROR_R_CHAR_P */
 
-	buf[buflen - 1] = '\0';
 	return (buf);
 } /* char *sstrerror */
 
@@ -329,28 +349,56 @@ int escape_slashes (char *buf, int buf_len)
 	return (0);
 } /* int escape_slashes */
 
-int timeval_sub_timespec (struct timeval *tv0, struct timeval *tv1, struct timespec *ret)
+int timeval_cmp (struct timeval tv0, struct timeval tv1, struct timeval *delta)
 {
-	if ((tv0 == NULL) || (tv1 == NULL) || (ret == NULL))
-		return (-2);
+	struct timeval *larger;
+	struct timeval *smaller;
 
-	if ((tv0->tv_sec < tv1->tv_sec)
-			|| ((tv0->tv_sec == tv1->tv_sec) && (tv0->tv_usec < tv1->tv_usec)))
-		return (-1);
+	int status;
 
-	ret->tv_sec  = tv0->tv_sec - tv1->tv_sec;
-	ret->tv_nsec = 1000 * ((long) (tv0->tv_usec - tv1->tv_usec));
+	NORMALIZE_TIMEVAL (tv0);
+	NORMALIZE_TIMEVAL (tv1);
 
-	if (ret->tv_nsec < 0)
+	if ((tv0.tv_sec == tv1.tv_sec) && (tv0.tv_usec == tv1.tv_usec))
 	{
-		assert (ret->tv_sec > 0);
-
-		ret->tv_nsec += 1000000000;
-		ret->tv_sec  -= 1;
+		if (delta != NULL) {
+			delta->tv_sec  = 0;
+			delta->tv_usec = 0;
+		}
+		return (0);
 	}
 
-	return (0);
-}
+	if ((tv0.tv_sec < tv1.tv_sec)
+			|| ((tv0.tv_sec == tv1.tv_sec) && (tv0.tv_usec < tv1.tv_usec)))
+	{
+		larger  = &tv1;
+		smaller = &tv0;
+		status  = -1;
+	}
+	else
+	{
+		larger  = &tv0;
+		smaller = &tv1;
+		status  = 1;
+	}
+
+	if (delta != NULL) {
+		delta->tv_sec = larger->tv_sec - smaller->tv_sec;
+
+		if (smaller->tv_usec <= larger->tv_usec)
+			delta->tv_usec = larger->tv_usec - smaller->tv_usec;
+		else
+		{
+			--delta->tv_sec;
+			delta->tv_usec = 1000000 + larger->tv_usec - smaller->tv_usec;
+		}
+	}
+
+	assert ((delta == NULL)
+			|| ((0 <= delta->tv_usec) && (delta->tv_usec < 1000000)));
+
+	return (status);
+} /* int timeval_cmp */
 
 int check_create_dir (const char *file_orig)
 {
@@ -365,7 +413,7 @@ int check_create_dir (const char *file_orig)
 	char *saveptr;
 	int   last_is_file = 1;
 	int   path_is_absolute = 0;
-	int   len;
+	size_t len;
 	int   i;
 
 	/*
@@ -376,7 +424,7 @@ int check_create_dir (const char *file_orig)
 
 	if ((len = strlen (file_orig)) < 1)
 		return (-1);
-	else if (len >= 512)
+	else if (len >= sizeof (file_copy))
 		return (-1);
 
 	/*
@@ -391,8 +439,7 @@ int check_create_dir (const char *file_orig)
 	/*
 	 * Create a copy for `strtok_r' to destroy
 	 */
-	strncpy (file_copy, file_orig, 512);
-	file_copy[511] = '\0';
+	sstrncpy (file_copy, file_orig, sizeof (file_copy));
 
 	/*
 	 * Break into components. This will eat up several slashes in a row and
@@ -422,7 +469,8 @@ int check_create_dir (const char *file_orig)
 		 */
 		if (fields[i][0] == '.')
 		{
-			ERROR ("Cowardly refusing to create a directory that begins with a `.' (dot): `%s'", file_orig);
+			ERROR ("Cowardly refusing to create a directory that "
+					"begins with a `.' (dot): `%s'", file_orig);
 			return (-2);
 		}
 
@@ -437,32 +485,42 @@ int check_create_dir (const char *file_orig)
 			return (-1);
 		}
 
-		if (stat (dir, &statbuf) == -1)
-		{
-			if (errno == ENOENT)
+		while (42) {
+			if (stat (dir, &statbuf) == -1)
 			{
-				if (mkdir (dir, 0755) == -1)
+				if (errno == ENOENT)
 				{
+					if (mkdir (dir, 0755) == 0)
+						break;
+
+					/* this might happen, if a different thread created
+					 * the directory in the meantime
+					 * => call stat() again to check for S_ISDIR() */
+					if (EEXIST == errno)
+						continue;
+
 					char errbuf[1024];
 					ERROR ("check_create_dir: mkdir (%s): %s", dir,
 							sstrerror (errno,
 								errbuf, sizeof (errbuf)));
 					return (-1);
 				}
+				else
+				{
+					char errbuf[1024];
+					ERROR ("check_create_dir: stat (%s): %s", dir,
+							sstrerror (errno, errbuf,
+								sizeof (errbuf)));
+					return (-1);
+				}
 			}
-			else
+			else if (!S_ISDIR (statbuf.st_mode))
 			{
-				char errbuf[1024];
-				ERROR ("stat (%s): %s", dir,
-						sstrerror (errno, errbuf,
-							sizeof (errbuf)));
+				ERROR ("check_create_dir: `%s' exists but is not "
+						"a directory!", dir);
 				return (-1);
 			}
-		}
-		else if (!S_ISDIR (statbuf.st_mode))
-		{
-			ERROR ("stat (%s): Not a directory!", dir);
-			return (-1);
+			break;
 		}
 	}
 
@@ -473,27 +531,26 @@ int check_create_dir (const char *file_orig)
 int get_kstat (kstat_t **ksp_ptr, char *module, int instance, char *name)
 {
 	char ident[128];
+
+	*ksp_ptr = NULL;
 	
 	if (kc == NULL)
 		return (-1);
 
-	snprintf (ident, 128, "%s,%i,%s", module, instance, name);
-	ident[127] = '\0';
+	ssnprintf (ident, sizeof (ident), "%s,%i,%s", module, instance, name);
 
+	*ksp_ptr = kstat_lookup (kc, module, instance, name);
 	if (*ksp_ptr == NULL)
 	{
-		if ((*ksp_ptr = kstat_lookup (kc, module, instance, name)) == NULL)
-		{
-			ERROR ("Cound not find kstat %s", ident);
-			return (-1);
-		}
+		ERROR ("get_kstat: Cound not find kstat %s", ident);
+		return (-1);
+	}
 
-		if ((*ksp_ptr)->ks_type != KSTAT_TYPE_NAMED)
-		{
-			WARNING ("kstat %s has wrong type", ident);
-			*ksp_ptr = NULL;
-			return (-1);
-		}
+	if ((*ksp_ptr)->ks_type != KSTAT_TYPE_NAMED)
+	{
+		ERROR ("get_kstat: kstat %s has wrong type", ident);
+		*ksp_ptr = NULL;
+		return (-1);
 	}
 
 #ifdef assert
@@ -503,13 +560,13 @@ int get_kstat (kstat_t **ksp_ptr, char *module, int instance, char *name)
 
 	if (kstat_read (kc, *ksp_ptr, NULL) == -1)
 	{
-		WARNING ("kstat %s could not be read", ident);
+		ERROR ("get_kstat: kstat %s could not be read", ident);
 		return (-1);
 	}
 
 	if ((*ksp_ptr)->ks_type != KSTAT_TYPE_NAMED)
 	{
-		WARNING ("kstat %s has wrong type", ident);
+		ERROR ("get_kstat: kstat %s has wrong type", ident);
 		return (-1);
 	}
 
@@ -527,12 +584,12 @@ long long get_kstat_value (kstat_t *ksp, char *name)
 #else
 	if (ksp == NULL)
 	{
-		fprintf (stderr, "ERROR: %s:%i: ksp == NULL\n", __FILE__, __LINE__);
+		ERROR ("ERROR: %s:%i: ksp == NULL\n", __FILE__, __LINE__);
 		return (-1LL);
 	}
 	else if (ksp->ks_type != KSTAT_TYPE_NAMED)
 	{
-		fprintf (stderr, "ERROR: %s:%i: ksp->ks_type != KSTAT_TYPE_NAMED\n", __FILE__, __LINE__);
+		ERROR ("ERROR: %s:%i: ksp->ks_type != KSTAT_TYPE_NAMED\n", __FILE__, __LINE__);
 		return (-1LL);
 	}
 #endif
@@ -663,21 +720,21 @@ int format_name (char *ret, int ret_len,
 	if ((plugin_instance == NULL) || (strlen (plugin_instance) == 0))
 	{
 		if ((type_instance == NULL) || (strlen (type_instance) == 0))
-			status = snprintf (ret, ret_len, "%s/%s/%s",
+			status = ssnprintf (ret, ret_len, "%s/%s/%s",
 					hostname, plugin, type);
 		else
-			status = snprintf (ret, ret_len, "%s/%s/%s-%s",
+			status = ssnprintf (ret, ret_len, "%s/%s/%s-%s",
 					hostname, plugin, type,
 					type_instance);
 	}
 	else
 	{
 		if ((type_instance == NULL) || (strlen (type_instance) == 0))
-			status = snprintf (ret, ret_len, "%s/%s-%s/%s",
+			status = ssnprintf (ret, ret_len, "%s/%s-%s/%s",
 					hostname, plugin, plugin_instance,
 					type);
 		else
-			status = snprintf (ret, ret_len, "%s/%s-%s/%s-%s",
+			status = ssnprintf (ret, ret_len, "%s/%s-%s/%s-%s",
 					hostname, plugin, plugin_instance,
 					type, type_instance);
 	}
@@ -838,26 +895,93 @@ int notification_init (notification_t *n, int severity, const char *message,
 	n->severity = severity;
 
 	if (message != NULL)
-		strncpy (n->message, message, sizeof (n->message));
+		sstrncpy (n->message, message, sizeof (n->message));
 	if (host != NULL)
-		strncpy (n->host, host, sizeof (n->host));
+		sstrncpy (n->host, host, sizeof (n->host));
 	if (plugin != NULL)
-		strncpy (n->plugin, plugin, sizeof (n->plugin));
+		sstrncpy (n->plugin, plugin, sizeof (n->plugin));
 	if (plugin_instance != NULL)
-		strncpy (n->plugin_instance, plugin_instance,
+		sstrncpy (n->plugin_instance, plugin_instance,
 				sizeof (n->plugin_instance));
 	if (type != NULL)
-		strncpy (n->type, type, sizeof (n->type));
+		sstrncpy (n->type, type, sizeof (n->type));
 	if (type_instance != NULL)
-		strncpy (n->type_instance, type_instance,
+		sstrncpy (n->type_instance, type_instance,
 				sizeof (n->type_instance));
 
-	n->message[sizeof (n->message) - 1] = '\0';
-	n->host[sizeof (n->host) - 1] = '\0';
-	n->plugin[sizeof (n->plugin) - 1] = '\0';
-	n->plugin_instance[sizeof (n->plugin_instance) - 1] = '\0';
-	n->type[sizeof (n->type) - 1] = '\0';
-	n->type_instance[sizeof (n->type_instance) - 1] = '\0';
-
 	return (0);
 } /* int notification_init */
+
+int walk_directory (const char *dir, dirwalk_callback_f callback,
+		void *user_data)
+{
+	struct dirent *ent;
+	DIR *dh;
+	int success;
+	int failure;
+
+	success = 0;
+	failure = 0;
+
+	if ((dh = opendir (dir)) == NULL)
+	{
+		char errbuf[1024];
+		ERROR ("walk_directory: Cannot open '%s': %s", dir,
+				sstrerror (errno, errbuf, sizeof (errbuf)));
+		return -1;
+	}
+
+	while ((ent = readdir (dh)) != NULL)
+	{
+		int status;
+
+		if (ent->d_name[0] == '.')
+			continue;
+
+		status = (*callback) (dir, ent->d_name, user_data);
+		if (status != 0)
+			failure++;
+		else
+			success++;
+	}
+
+	closedir (dh);
+
+	if ((success == 0) && (failure > 0))
+		return (-1);
+	return (0);
+}
+
+int read_file_contents (const char *filename, char *buf, int bufsize)
+{
+	FILE *fh;
+	int n;
+
+	if ((fh = fopen (filename, "r")) == NULL)
+		return -1;
+
+	n = fread(buf, 1, bufsize, fh);
+	fclose(fh);
+
+	return n;
+}
+
+counter_t counter_diff (counter_t old_value, counter_t new_value)
+{
+	counter_t diff;
+
+	if (old_value > new_value)
+	{
+		if (old_value <= 4294967295U)
+			diff = (4294967295U - old_value) + new_value;
+		else
+			diff = (18446744073709551615ULL - old_value)
+				+ new_value;
+	}
+	else
+	{
+		diff = new_value - old_value;
+	}
+
+	return (diff);
+} /* counter_t counter_to_gauge */
