@@ -124,13 +124,13 @@ static void df_submit (char *df_name,
 
 	vl.values = values;
 	vl.values_len = 2;
-	vl.time = time (NULL);
 	sstrncpy (vl.host, hostname_g, sizeof (vl.host));
 	sstrncpy (vl.plugin, "df", sizeof (vl.plugin));
 	sstrncpy (vl.plugin_instance, "", sizeof (vl.plugin_instance));
-	strncpy (vl.type_instance, df_name, sizeof (vl.type_instance));
+	sstrncpy (vl.type, "df", sizeof (vl.host));
+	sstrncpy (vl.type_instance, df_name, sizeof (vl.type_instance));
 
-	plugin_dispatch_values ("df", &vl);
+	plugin_dispatch_values (&vl);
 } /* void df_submit */
 
 static int df_read (void)
@@ -173,13 +173,13 @@ static int df_read (void)
 
 		if (strcmp (mnt_ptr->dir, "/") == 0)
 		{
-			strncpy (mnt_name, "root", sizeof (mnt_name));
+			sstrncpy (mnt_name, "root", sizeof (mnt_name));
 		}
 		else
 		{
 			int i, len;
 
-			strncpy (mnt_name, mnt_ptr->dir + 1, sizeof (mnt_name));
+			sstrncpy (mnt_name, mnt_ptr->dir + 1, sizeof (mnt_name));
 			len = strlen (mnt_name);
 
 			for (i = 0; i < len; i++)
