@@ -212,13 +212,13 @@ static void disk_submit (const char *plugin_instance,
 
 	vl.values = values;
 	vl.values_len = 2;
-	vl.time = time (NULL);
 	sstrncpy (vl.host, hostname_g, sizeof (vl.host));
 	sstrncpy (vl.plugin, "disk", sizeof (vl.plugin));
-	strncpy (vl.plugin_instance, plugin_instance,
+	sstrncpy (vl.plugin_instance, plugin_instance,
 			sizeof (vl.plugin_instance));
+	sstrncpy (vl.type, type, sizeof (vl.type));
 
-	plugin_dispatch_values (type, &vl);
+	plugin_dispatch_values (&vl);
 } /* void disk_submit */
 
 #if HAVE_IOKIT_IOKITLIB_H
@@ -375,7 +375,8 @@ static int disk_read (void)
 		write_tme = dict_get_value (stats_dict,
 				kIOBlockStorageDriverStatisticsTotalWriteTimeKey);
 
-		if (snprintf (disk_name, 64, "%i-%i", disk_major, disk_minor) >= 64)
+		if (ssnprintf (disk_name, sizeof (disk_name),
+				"%i-%i", disk_major, disk_minor) >= sizeof (disk_name))
 		{
 			DEBUG ("snprintf (major, minor) failed.");
 			CFRelease (child_dict);
