@@ -187,12 +187,18 @@ static int email_config (const char *key, const char *value)
 			fprintf (stderr, "email plugin: `MaxConns' was set to invalid "
 					"value %li, will use default %i.\n",
 					tmp, MAX_CONNS);
+			ERROR ("email plugin: `MaxConns' was set to invalid "
+					"value %li, will use default %i.\n",
+					tmp, MAX_CONNS);
 			max_conns = MAX_CONNS;
 		}
 		else if (tmp > MAX_CONNS_LIMIT) {
 			fprintf (stderr, "email plugin: `MaxConns' was set to invalid "
 					"value %li, will use hardcoded limit %i.\n",
 					tmp, MAX_CONNS_LIMIT);
+			ERROR ("email plugin: `MaxConns' was set to invalid "
+					"value %li, will use hardcoded limit %i.\n",
+					tmp, MAX_CONNS_LIMIT);
 			max_conns = MAX_CONNS_LIMIT;
 		}
 		else {
@@ -354,8 +360,8 @@ static void *collect (void *arg)
 			}
 		} /* while (loop) */
 
-		log_debug ("[thread #%5lu] shutting down connection on fd #%i",
-				pthread_self (), fileno (this->socket));
+		log_debug ("Shutting down connection on fd #%i",
+				fileno (this->socket));
 
 		fclose (connection->socket);
 		free (connection);
@@ -372,7 +378,7 @@ static void *collect (void *arg)
 	pthread_exit ((void *)0);
 } /* static void *collect (void *) */
 
-static void *open_connection (void *arg)
+static void *open_connection (void __attribute__((unused)) *arg)
 {
 	struct sockaddr_un addr;
 
@@ -390,9 +396,7 @@ static void *open_connection (void *arg)
 	}
 
 	addr.sun_family = AF_UNIX;
-
-	strncpy (addr.sun_path, path, (size_t)(UNIX_PATH_MAX - 1));
-	addr.sun_path[UNIX_PATH_MAX - 1] = '\0';
+	sstrncpy (addr.sun_path, path, (size_t)(UNIX_PATH_MAX - 1));
 
 	errno = 0;
 	if (-1 == bind (connector_socket, (struct sockaddr *)&addr,
@@ -652,12 +656,12 @@ static void email_submit (const char *type, const char *type_instance, gauge_t v
 
 	vl.values = values;
 	vl.values_len = 1;
-	vl.time = time (NULL);
 	sstrncpy (vl.host, hostname_g, sizeof (vl.host));
 	sstrncpy (vl.plugin, "email", sizeof (vl.plugin));
-	strncpy (vl.type_instance, type_instance, sizeof (vl.type_instance));
+	sstrncpy (vl.type, type, sizeof (vl.type));
+	sstrncpy (vl.type_instance, type_instance, sizeof (vl.type_instance));
 
-	plugin_dispatch_values (type, &vl);
+	plugin_dispatch_values (&vl);
 } /* void email_submit */
 
 /* Copy list l1 to list l2. l2 may partly exist already, but it is assumed
