@@ -1,6 +1,8 @@
 /**
  * collectd - src/hddtemp.c
  * Copyright (C) 2005,2006  Vincent Stehl
+ * Copyright (C) 2006,2007  Florian octo Forster
+ * Copyright (C) 2008       Sebastian Harl
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -19,6 +21,7 @@
  * Authors:
  *   Vincent Stehl <vincent.stehle at free.fr>
  *   Florian octo Forster <octo at verplant.org>
+ *   Sebastian Harl <sh at tokkee.org>
  *
  * TODO:
  *   Do a pass, some day, and spare some memory. We consume too much for now
@@ -223,11 +226,10 @@ static int hddtemp_config (const char *key, const char *value)
 	{
 		int port = (int) (atof (value));
 		if ((port > 0) && (port <= 65535))
-			snprintf (hddtemp_port, sizeof (hddtemp_port),
+			ssnprintf (hddtemp_port, sizeof (hddtemp_port),
 					"%i", port);
 		else
-			strncpy (hddtemp_port, value, sizeof (hddtemp_port));
-		hddtemp_port[sizeof (hddtemp_port) - 1] = '\0';
+			sstrncpy (hddtemp_port, value, sizeof (hddtemp_port));
 	}
 	else if (strcasecmp (key, "TranslateDevicename") == 0)
 	{
@@ -430,7 +432,7 @@ static char *hddtemp_get_major_minor (char *drive)
 	if ((ret = (char *) malloc (128 * sizeof (char))) == NULL)
 		return (NULL);
 
-	if (snprintf (ret, 128, "%i-%i", list->major, list->minor) >= 128)
+	if (ssnprintf (ret, 128, "%i-%i", list->major, list->minor) >= 128)
 	{
 		free (ret);
 		return (NULL);
@@ -448,12 +450,12 @@ static void hddtemp_submit (char *type_instance, double value)
 
 	vl.values = values;
 	vl.values_len = 1;
-	vl.time = time (NULL);
 	sstrncpy (vl.host, hostname_g, sizeof (vl.host));
 	sstrncpy (vl.plugin, "hddtemp", sizeof (vl.plugin));
-	strncpy (vl.type_instance, type_instance, sizeof (vl.type_instance));
+	sstrncpy (vl.type, "temperature", sizeof (vl.type));
+	sstrncpy (vl.type_instance, type_instance, sizeof (vl.type_instance));
 
-	plugin_dispatch_values ("temperature", &vl);
+	plugin_dispatch_values (&vl);
 }
 
 static int hddtemp_read (void)
