@@ -1,6 +1,6 @@
 /**
  * collectd - src/load.c
- * Copyright (C) 2005-2007  Florian octo Forster
+ * Copyright (C) 2005-2008  Florian octo Forster
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -50,11 +50,11 @@ static void load_submit (gauge_t snum, gauge_t mnum, gauge_t lnum)
 
 	vl.values = values;
 	vl.values_len = STATIC_ARRAY_SIZE (values);
-	vl.time = time (NULL);
 	sstrncpy (vl.host, hostname_g, sizeof (vl.host));
 	sstrncpy (vl.plugin, "load", sizeof (vl.plugin));
+	sstrncpy (vl.type, "load", sizeof (vl.type));
 
-	plugin_dispatch_values ("load", &vl);
+	plugin_dispatch_values (&vl);
 }
 
 static int load_read (void)
