@@ -1,6 +1,6 @@
 /**
  * collectd - src/memory.c
- * Copyright (C) 2005-2007  Florian octo Forster
+ * Copyright (C) 2005-2008  Florian octo Forster
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -92,8 +92,11 @@ static int memory_init (void)
 #elif defined(HAVE_LIBKSTAT)
 	/* getpagesize(3C) tells me this does not fail.. */
 	pagesize = getpagesize ();
-	if (get_kstat (&ksp, "unix", 0, "system_pages"))
+	if (get_kstat (&ksp, "unix", 0, "system_pages") != 0)
+	{
 		ksp = NULL;
+		return (-1);
+	}
 #endif /* HAVE_LIBKSTAT */
 
 	return (0);
@@ -108,13 +111,12 @@ static void memory_submit (const char *type_instance, gauge_t value)
 
 	vl.values = values;
 	vl.values_len = 1;
-	vl.time = time (NULL);
 	sstrncpy (vl.host, hostname_g, sizeof (vl.host));
 	sstrncpy (vl.plugin, "memory", sizeof (vl.plugin));
-	strncpy (vl.type_instance, type_instance, sizeof (vl.type_instance));
-	vl.type_instance[sizeof (vl.type_instance) - 1] = '\0';
+	sstrncpy (vl.type, "memory", sizeof (vl.type));
+	sstrncpy (vl.type_instance, type_instance, sizeof (vl.type_instance));
 
-	plugin_dispatch_values ("memory", &vl);
+	plugin_dispatch_values (&vl);
 }
 
 static int memory_read (void)
@@ -206,7 +208,7 @@ static int memory_read (void)
 					NULL, 0) == 0)
 		{
 			sysctl_vals[i] = value;
-			DEBUG ("memory plugin: %26s: %6i", sysctl_keys[i], sysctl_vals[i]);
+			DEBUG ("memory plugin: %26s: %g", sysctl_keys[i], sysctl_vals[i]);
 		}
 		else
 		{
@@ -319,7 +321,7 @@ static int memory_read (void)
 	if ((ios = sg_get_mem_stats ()) != NULL)
 	{
 		memory_submit ("used",   ios->used);
-		memory_submit ("cached", ios->cached);
+		memory_submit ("cached", ios->cache);
 		memory_submit ("free",   ios->free);
 	}
 #endif /* HAVE_LIBSTATGRAB */
