@@ -1,6 +1,6 @@
 /**
  * collectd - src/rrdtool.c
- * Copyright (C) 2006,2007  Florian octo Forster
+ * Copyright (C) 2006-2008  Florian octo Forster
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -23,6 +23,7 @@
 #include "plugin.h"
 #include "common.h"
 #include "utils_avltree.h"
+#include "utils_rrdcreate.h"
 
 #include <rrd.h>
 
@@ -42,11 +43,19 @@ struct rrd_cache_s
 	enum
 	{
 		FLAG_NONE   = 0x00,
-		FLAG_QUEUED = 0x01
+		FLAG_QUEUED = 0x01,
+		FLAG_FLUSHQ = 0x02
 	} flags;
 };
 typedef struct rrd_cache_s rrd_cache_t;
 
+enum rrd_queue_dir_e
+{
+  QUEUE_INSERT_FRONT,
+  QUEUE_INSERT_BACK
+};
+typedef enum rrd_queue_dir_e rrd_queue_dir_t;
+
 struct rrd_queue_s
 {
 	char *filename;
@@ -57,27 +66,6 @@ typedef struct rrd_queue_s rrd_queue_t;
 /*
  * Private variables
  */
-static int rra_timespans[] =
-{
-	3600,
-	86400,
-	604800,
-	2678400,
-	31622400
-};
-static int rra_timespans_num = STATIC_ARRAY_SIZE (rra_timespans);
-
-static int *rra_timespans_custom = NULL;
-static int rra_timespans_custom_num = 0;
-
-static char *rra_types[] =
-{
-	"AVERAGE",
-	"MIN",
-	"MAX"
-};
-static int rra_types_num = STATIC_ARRAY_SIZE (rra_types);
-
 static const char *config_keys[] =
 {
 	"CacheTimeout",
@@ -87,18 +75,29 @@ static const char *config_keys[] =
 	"HeartBeat",
 	"RRARows",
 	"RRATimespan",
-	"XFF"
+	"XFF",
+	"WritesPerSecond"
 };
 static int config_keys_num = STATIC_ARRAY_SIZE (config_keys);
 
 /* If datadir is zero, the daemon's basedir is used. If stepsize or heartbeat
  * is zero a default, depending on the `interval' member of the value list is
  * being used. */
-static char   *datadir   = NULL;
-static int     stepsize  = 0;
-static int     heartbeat = 0;
-static int     rrarows   = 1200;
-static double  xff       = 0.1;
+static char *datadir   = NULL;
+static double write_rate = 0.0;
+static rrdcreate_config_t rrdcreate_config =
+{
+	/* stepsize = */ 0,
+	/* heartbeat = */ 0,
+	/* rrarows = */ 1200,
+	/* xff = */ 0.1,
+
+	/* timespans = */ NULL,
+	/* timespans_num = */ 0,
+
+	/* consolidation_functions = */ NULL,
+	/* consolidation_functions_num = */ 0
+};
 
 /* XXX: If you need to lock both, cache_lock and queue_lock, at the same time,
  * ALWAYS lock `cache_lock' first! */
@@ -110,6 +109,8 @@ static pthread_mutex_t cache_lock = PTHREAD_MUTEX_INITIALIZER;
 
 static rrd_queue_t    *queue_head = NULL;
 static rrd_queue_t    *queue_tail = NULL;
+static rrd_queue_t    *flushq_head = NULL;
+static rrd_queue_t    *flushq_tail = NULL;
 static pthread_t       queue_thread = 0;
 static pthread_mutex_t queue_lock = PTHREAD_MUTEX_INITIALIZER;
 static pthread_cond_t  queue_cond = PTHREAD_COND_INITIALIZER;
@@ -120,240 +121,9 @@ static pthread_mutex_t librrd_lock = PTHREAD_MUTEX_INITIALIZER;
 
 static int do_shutdown = 0;
 
-/* * * * * * * * * *
- * WARNING:  Magic *
- * * * * * * * * * */
-
-static void rra_free (int rra_num, char **rra_def)
-{
-	int i;
-
-	for (i = 0; i < rra_num; i++)
-	{
-		sfree (rra_def[i]);
-	}
-	sfree (rra_def);
-} /* void rra_free */
-
-static int rra_get (char ***ret, const value_list_t *vl)
-{
-	char **rra_def;
-	int rra_num;
-
-	int *rts;
-	int  rts_num;
-
-	int rra_max;
-
-	int span;
-
-	int cdp_num;
-	int cdp_len;
-	int i, j;
-
-	char buffer[64];
-
-	/* The stepsize we use here: If it is user-set, use it. If not, use the
-	 * interval of the value-list. */
-	int ss;
-
-	if (rrarows <= 0)
-	{
-		*ret = NULL;
-		return (-1);
-	}
-
-	ss = (stepsize > 0) ? stepsize : vl->interval;
-	if (ss <= 0)
-	{
-		*ret = NULL;
-		return (-1);
-	}
-
-	/* Use the configured timespans or fall back to the built-in defaults */
-	if (rra_timespans_custom_num != 0)
-	{
-		rts = rra_timespans_custom;
-		rts_num = rra_timespans_custom_num;
-	}
-	else
-	{
-		rts = rra_timespans;
-		rts_num = rra_timespans_num;
-	}
-
-	rra_max = rts_num * rra_types_num;
-
-	if ((rra_def = (char **) malloc ((rra_max + 1) * sizeof (char *))) == NULL)
-		return (-1);
-	memset (rra_def, '\0', (rra_max + 1) * sizeof (char *));
-	rra_num = 0;
-
-	cdp_len = 0;
-	for (i = 0; i < rts_num; i++)
-	{
-		span = rts[i];
-
-		if ((span / ss) < rrarows)
-			span = ss * rrarows;
-
-		if (cdp_len == 0)
-			cdp_len = 1;
-		else
-			cdp_len = (int) floor (((double) span)
-					/ ((double) (rrarows * ss)));
-
-		cdp_num = (int) ceil (((double) span)
-				/ ((double) (cdp_len * ss)));
-
-		for (j = 0; j < rra_types_num; j++)
-		{
-			if (rra_num >= rra_max)
-				break;
-
-			if (snprintf (buffer, sizeof (buffer), "RRA:%s:%3.1f:%u:%u",
-						rra_types[j], xff,
-						cdp_len, cdp_num) >= sizeof (buffer))
-			{
-				ERROR ("rra_get: Buffer would have been truncated.");
-				continue;
-			}
-
-			rra_def[rra_num++] = sstrdup (buffer);
-		}
-	}
-
-#if COLLECT_DEBUG
-	DEBUG ("rra_num = %i", rra_num);
-	for (i = 0; i < rra_num; i++)
-		DEBUG ("  %s", rra_def[i]);
-#endif
-
-	*ret = rra_def;
-	return (rra_num);
-} /* int rra_get */
-
-static void ds_free (int ds_num, char **ds_def)
-{
-	int i;
-
-	for (i = 0; i < ds_num; i++)
-		if (ds_def[i] != NULL)
-			free (ds_def[i]);
-	free (ds_def);
-}
-
-static int ds_get (char ***ret, const data_set_t *ds, const value_list_t *vl)
-{
-	char **ds_def;
-	int ds_num;
-
-	char min[32];
-	char max[32];
-	char buffer[128];
-
-	DEBUG ("ds->ds_num = %i", ds->ds_num);
-
-	ds_def = (char **) malloc (ds->ds_num * sizeof (char *));
-	if (ds_def == NULL)
-	{
-		char errbuf[1024];
-		ERROR ("rrdtool plugin: malloc failed: %s",
-				sstrerror (errno, errbuf, sizeof (errbuf)));
-		return (-1);
-	}
-	memset (ds_def, '\0', ds->ds_num * sizeof (char *));
-
-	for (ds_num = 0; ds_num < ds->ds_num; ds_num++)
-	{
-		data_source_t *d = ds->ds + ds_num;
-		char *type;
-		int status;
-
-		ds_def[ds_num] = NULL;
-
-		if (d->type == DS_TYPE_COUNTER)
-			type = "COUNTER";
-		else if (d->type == DS_TYPE_GAUGE)
-			type = "GAUGE";
-		else
-		{
-			ERROR ("rrdtool plugin: Unknown DS type: %i",
-					d->type);
-			break;
-		}
-
-		if (isnan (d->min))
-		{
-			sstrncpy (min, "U", sizeof (min));
-		}
-		else
-		{
-			snprintf (min, sizeof (min), "%lf", d->min);
-			min[sizeof (min) - 1] = '\0';
-		}
-
-		if (isnan (d->max))
-		{
-			sstrncpy (max, "U", sizeof (max));
-		}
-		else
-		{
-			snprintf (max, sizeof (max), "%lf", d->max);
-			max[sizeof (max) - 1] = '\0';
-		}
-
-		status = snprintf (buffer, sizeof (buffer),
-				"DS:%s:%s:%i:%s:%s",
-				d->name, type,
-				(heartbeat > 0) ? heartbeat : (2 * vl->interval),
-				min, max);
-		if ((status < 1) || (status >= sizeof (buffer)))
-			break;
-
-		ds_def[ds_num] = sstrdup (buffer);
-	} /* for ds_num = 0 .. ds->ds_num */
-
-#if COLLECT_DEBUG
-{
-	int i;
-	DEBUG ("ds_num = %i", ds_num);
-	for (i = 0; i < ds_num; i++)
-		DEBUG ("  %s", ds_def[i]);
-}
-#endif
-
-	if (ds_num != ds->ds_num)
-	{
-		ds_free (ds_num, ds_def);
-		return (-1);
-	}
-
-	*ret = ds_def;
-	return (ds_num);
-}
-
 #if HAVE_THREADSAFE_LIBRRD
-static int srrd_create (char *filename, unsigned long pdp_step, time_t last_up,
-		int argc, char **argv)
-{
-	int status;
-
-	optind = 0; /* bug in librrd? */
-	rrd_clear_error ();
-
-	status = rrd_create_r (filename, pdp_step, last_up, argc, (void *) argv);
-
-	if (status != 0)
-	{
-		WARNING ("rrdtool plugin: rrd_create_r (%s) failed: %s",
-				filename, rrd_get_error ());
-	}
-
-	return (status);
-} /* int srrd_create */
-
-static int srrd_update (char *filename, char *template, int argc, char **argv)
+static int srrd_update (char *filename, char *template,
+		int argc, const char **argv)
 {
 	int status;
 
@@ -373,62 +143,8 @@ static int srrd_update (char *filename, char *template, int argc, char **argv)
 /* #endif HAVE_THREADSAFE_LIBRRD */
 
 #else /* !HAVE_THREADSAFE_LIBRRD */
-static int srrd_create (char *filename, unsigned long pdp_step, time_t last_up,
-		int argc, char **argv)
-{
-	int status;
-
-	int new_argc;
-	char **new_argv;
-
-	char pdp_step_str[16];
-	char last_up_str[16];
-
-	new_argc = 6 + argc;
-	new_argv = (char **) malloc ((new_argc + 1) * sizeof (char *));
-	if (new_argv == NULL)
-	{
-		ERROR ("rrdtool plugin: malloc failed.");
-		return (-1);
-	}
-
-	if (last_up == 0)
-		last_up = time (NULL) - 10;
-
-	snprintf (pdp_step_str, sizeof (pdp_step_str), "%lu", pdp_step);
-	pdp_step_str[sizeof (pdp_step_str) - 1] = '\0';
-	snprintf (last_up_str, sizeof (last_up_str), "%u", (unsigned int) last_up);
-	last_up_str[sizeof (last_up_str) - 1] = '\0';
-
-	new_argv[0] = "create";
-	new_argv[1] = filename;
-	new_argv[2] = "-s";
-	new_argv[3] = pdp_step_str;
-	new_argv[4] = "-b";
-	new_argv[5] = last_up_str;
-
-	memcpy (new_argv + 6, argv, argc * sizeof (char *));
-	new_argv[new_argc] = NULL;
-	
-	pthread_mutex_lock (&librrd_lock);
-	optind = 0; /* bug in librrd? */
-	rrd_clear_error ();
-
-	status = rrd_create (new_argc, new_argv);
-	pthread_mutex_unlock (&librrd_lock);
-
-	if (status != 0)
-	{
-		WARNING ("rrdtool plugin: rrd_create (%s) failed: %s",
-				filename, rrd_get_error ());
-	}
-
-	sfree (new_argv);
-
-	return (status);
-} /* int srrd_create */
-
-static int srrd_update (char *filename, char *template, int argc, char **argv)
+static int srrd_update (char *filename, char *template,
+		int argc, const char **argv)
 {
 	int status;
 
@@ -470,58 +186,6 @@ static int srrd_update (char *filename, char *template, int argc, char **argv)
 } /* int srrd_update */
 #endif /* !HAVE_THREADSAFE_LIBRRD */
 
-static int rrd_create_file (char *filename, const data_set_t *ds, const value_list_t *vl)
-{
-	char **argv;
-	int argc;
-	char **rra_def;
-	int rra_num;
-	char **ds_def;
-	int ds_num;
-	int status = 0;
-
-	if (check_create_dir (filename))
-		return (-1);
-
-	if ((rra_num = rra_get (&rra_def, vl)) < 1)
-	{
-		ERROR ("rrd_create_file failed: Could not calculate RRAs");
-		return (-1);
-	}
-
-	if ((ds_num = ds_get (&ds_def, ds, vl)) < 1)
-	{
-		ERROR ("rrd_create_file failed: Could not calculate DSes");
-		return (-1);
-	}
-
-	argc = ds_num + rra_num;
-
-	if ((argv = (char **) malloc (sizeof (char *) * (argc + 1))) == NULL)
-	{
-		char errbuf[1024];
-		ERROR ("rrd_create failed: %s",
-				sstrerror (errno, errbuf, sizeof (errbuf)));
-		return (-1);
-	}
-
-	memcpy (argv, ds_def, ds_num * sizeof (char *));
-	memcpy (argv + ds_num, rra_def, rra_num * sizeof (char *));
-	argv[ds_num + rra_num] = NULL;
-
-	assert (vl->time > 10);
-	status = srrd_create (filename,
-			(stepsize > 0) ? stepsize : vl->interval,
-			vl->time - 10,
-			argc, argv);
-
-	free (argv);
-	ds_free (ds_num, ds_def);
-	rra_free (rra_num, rra_def);
-
-	return (status);
-}
-
 static int value_list_to_string (char *buffer, int buffer_len,
 		const data_set_t *ds, const value_list_t *vl)
 {
@@ -531,7 +195,7 @@ static int value_list_to_string (char *buffer, int buffer_len,
 
 	memset (buffer, '\0', buffer_len);
 
-	status = snprintf (buffer, buffer_len, "%u", (unsigned int) vl->time);
+	status = ssnprintf (buffer, buffer_len, "%u", (unsigned int) vl->time);
 	if ((status < 1) || (status >= buffer_len))
 		return (-1);
 	offset = status;
@@ -543,10 +207,10 @@ static int value_list_to_string (char *buffer, int buffer_len,
 			return (-1);
 
 		if (ds->ds[i].type == DS_TYPE_COUNTER)
-			status = snprintf (buffer + offset, buffer_len - offset,
+			status = ssnprintf (buffer + offset, buffer_len - offset,
 					":%llu", vl->values[i].counter);
 		else
-			status = snprintf (buffer + offset, buffer_len - offset,
+			status = ssnprintf (buffer + offset, buffer_len - offset,
 					":%lf", vl->values[i].gauge);
 
 		if ((status < 1) || (status >= (buffer_len - offset)))
@@ -559,42 +223,42 @@ static int value_list_to_string (char *buffer, int buffer_len,
 } /* int value_list_to_string */
 
 static int value_list_to_filename (char *buffer, int buffer_len,
-		const data_set_t *ds, const value_list_t *vl)
+		const data_set_t __attribute__((unused)) *ds, const value_list_t *vl)
 {
 	int offset = 0;
 	int status;
 
 	if (datadir != NULL)
 	{
-		status = snprintf (buffer + offset, buffer_len - offset,
+		status = ssnprintf (buffer + offset, buffer_len - offset,
 				"%s/", datadir);
 		if ((status < 1) || (status >= buffer_len - offset))
 			return (-1);
 		offset += status;
 	}
 
-	status = snprintf (buffer + offset, buffer_len - offset,
+	status = ssnprintf (buffer + offset, buffer_len - offset,
 			"%s/", vl->host);
 	if ((status < 1) || (status >= buffer_len - offset))
 		return (-1);
 	offset += status;
 
 	if (strlen (vl->plugin_instance) > 0)
-		status = snprintf (buffer + offset, buffer_len - offset,
+		status = ssnprintf (buffer + offset, buffer_len - offset,
 				"%s-%s/", vl->plugin, vl->plugin_instance);
 	else
-		status = snprintf (buffer + offset, buffer_len - offset,
+		status = ssnprintf (buffer + offset, buffer_len - offset,
 				"%s/", vl->plugin);
 	if ((status < 1) || (status >= buffer_len - offset))
 		return (-1);
 	offset += status;
 
 	if (strlen (vl->type_instance) > 0)
-		status = snprintf (buffer + offset, buffer_len - offset,
-				"%s-%s.rrd", ds->type, vl->type_instance);
+		status = ssnprintf (buffer + offset, buffer_len - offset,
+				"%s-%s.rrd", vl->type, vl->type_instance);
 	else
-		status = snprintf (buffer + offset, buffer_len - offset,
-				"%s.rrd", ds->type);
+		status = ssnprintf (buffer + offset, buffer_len - offset,
+				"%s.rrd", vl->type);
 	if ((status < 1) || (status >= buffer_len - offset))
 		return (-1);
 	offset += status;
@@ -602,37 +266,96 @@ static int value_list_to_filename (char *buffer, int buffer_len,
 	return (0);
 } /* int value_list_to_filename */
 
-static void *rrd_queue_thread (void *data)
+static void *rrd_queue_thread (void __attribute__((unused)) *data)
 {
+        struct timeval tv_next_update;
+        struct timeval tv_now;
+
+        gettimeofday (&tv_next_update, /* timezone = */ NULL);
+
 	while (42)
 	{
 		rrd_queue_t *queue_entry;
 		rrd_cache_t *cache_entry;
 		char **values;
 		int    values_num;
+		int    status;
 		int    i;
 
-		/* XXX: If you need to lock both, cache_lock and queue_lock, at
-		 * the same time, ALWAYS lock `cache_lock' first! */
+		values = NULL;
+		values_num = 0;
 
-		/* wait until an entry is available */
-		pthread_mutex_lock (&queue_lock);
-		while ((queue_head == NULL) && (do_shutdown == 0))
-			pthread_cond_wait (&queue_cond, &queue_lock);
+                pthread_mutex_lock (&queue_lock);
+                /* Wait for values to arrive */
+                while (true)
+                {
+                  struct timespec ts_wait;
 
-		/* We're in the shutdown phase */
-		if (queue_head == NULL)
-		{
-			pthread_mutex_unlock (&queue_lock);
-			break;
-		}
+                  while ((flushq_head == NULL) && (queue_head == NULL)
+                      && (do_shutdown == 0))
+                    pthread_cond_wait (&queue_cond, &queue_lock);
 
-		/* Dequeue the first entry */
-		queue_entry = queue_head;
-		if (queue_head == queue_tail)
-			queue_head = queue_tail = NULL;
-		else
-			queue_head = queue_head->next;
+                  if ((flushq_head == NULL) && (queue_head == NULL))
+                    break;
+
+                  /* Don't delay if there's something to flush */
+                  if (flushq_head != NULL)
+                    break;
+
+                  /* Don't delay if we're shutting down */
+                  if (do_shutdown != 0)
+                    break;
+
+                  /* Don't delay if no delay was configured. */
+                  if (write_rate <= 0.0)
+                    break;
+
+                  gettimeofday (&tv_now, /* timezone = */ NULL);
+                  status = timeval_cmp (tv_next_update, tv_now, NULL);
+                  /* We're good to go */
+                  if (status <= 0)
+                    break;
+
+                  /* We're supposed to wait a bit with this update, so we'll
+                   * wait for the next addition to the queue or to the end of
+                   * the wait period - whichever comes first. */
+                  ts_wait.tv_sec = tv_next_update.tv_sec;
+                  ts_wait.tv_nsec = 1000 * tv_next_update.tv_usec;
+
+                  status = pthread_cond_timedwait (&queue_cond, &queue_lock,
+                      &ts_wait);
+                  if (status == ETIMEDOUT)
+                    break;
+                } /* while (true) */
+
+                /* XXX: If you need to lock both, cache_lock and queue_lock, at
+                 * the same time, ALWAYS lock `cache_lock' first! */
+
+                /* We're in the shutdown phase */
+                if ((flushq_head == NULL) && (queue_head == NULL))
+                {
+                  pthread_mutex_unlock (&queue_lock);
+                  break;
+                }
+
+                if (flushq_head != NULL)
+                {
+                  /* Dequeue the first flush entry */
+                  queue_entry = flushq_head;
+                  if (flushq_head == flushq_tail)
+                    flushq_head = flushq_tail = NULL;
+                  else
+                    flushq_head = flushq_head->next;
+                }
+                else /* if (queue_head != NULL) */
+                {
+                  /* Dequeue the first regular entry */
+                  queue_entry = queue_head;
+                  if (queue_head == queue_tail)
+                    queue_head = queue_tail = NULL;
+                  else
+                    queue_head = queue_head->next;
+                }
 
 		/* Unlock the queue again */
 		pthread_mutex_unlock (&queue_lock);
@@ -641,19 +364,45 @@ static void *rrd_queue_thread (void *data)
 		 * we make a copy of it's values */
 		pthread_mutex_lock (&cache_lock);
 
-		c_avl_get (cache, queue_entry->filename, (void *) &cache_entry);
+		status = c_avl_get (cache, queue_entry->filename,
+				(void *) &cache_entry);
 
-		values = cache_entry->values;
-		values_num = cache_entry->values_num;
+		if (status == 0)
+		{
+			values = cache_entry->values;
+			values_num = cache_entry->values_num;
 
-		cache_entry->values = NULL;
-		cache_entry->values_num = 0;
-		cache_entry->flags = FLAG_NONE;
+			cache_entry->values = NULL;
+			cache_entry->values_num = 0;
+			cache_entry->flags = FLAG_NONE;
+		}
 
 		pthread_mutex_unlock (&cache_lock);
 
+		if (status != 0)
+		{
+			sfree (queue_entry->filename);
+			sfree (queue_entry);
+			continue;
+		}
+
+		/* Update `tv_next_update' */
+		if (write_rate > 0.0) 
+                {
+                  gettimeofday (&tv_now, /* timezone = */ NULL);
+                  tv_next_update.tv_sec = tv_now.tv_sec;
+                  tv_next_update.tv_usec = tv_now.tv_usec
+                    + ((suseconds_t) (1000000 * write_rate));
+                  while (tv_next_update.tv_usec > 1000000)
+                  {
+                    tv_next_update.tv_sec++;
+                    tv_next_update.tv_usec -= 1000000;
+                  }
+                }
+
 		/* Write the values to the RRD-file */
-		srrd_update (queue_entry->filename, NULL, values_num, values);
+		srrd_update (queue_entry->filename, NULL,
+				values_num, (const char **)values);
 		DEBUG ("rrdtool plugin: queue thread: Wrote %i values to %s",
 				values_num, queue_entry->filename);
 
@@ -675,36 +424,79 @@ static void *rrd_queue_thread (void *data)
 	return ((void *) 0);
 } /* void *rrd_queue_thread */
 
-static int rrd_queue_cache_entry (const char *filename)
+static int rrd_queue_enqueue (const char *filename,
+    rrd_queue_t **head, rrd_queue_t **tail)
 {
-	rrd_queue_t *queue_entry;
+  rrd_queue_t *queue_entry;
 
-	queue_entry = (rrd_queue_t *) malloc (sizeof (rrd_queue_t));
-	if (queue_entry == NULL)
-		return (-1);
+  queue_entry = (rrd_queue_t *) malloc (sizeof (rrd_queue_t));
+  if (queue_entry == NULL)
+    return (-1);
 
-	queue_entry->filename = strdup (filename);
-	if (queue_entry->filename == NULL)
-	{
-		free (queue_entry);
-		return (-1);
-	}
+  queue_entry->filename = strdup (filename);
+  if (queue_entry->filename == NULL)
+  {
+    free (queue_entry);
+    return (-1);
+  }
 
-	queue_entry->next = NULL;
+  queue_entry->next = NULL;
 
-	pthread_mutex_lock (&queue_lock);
-	if (queue_tail == NULL)
-		queue_head = queue_entry;
-	else
-		queue_tail->next = queue_entry;
-	queue_tail = queue_entry;
-	pthread_cond_signal (&queue_cond);
-	pthread_mutex_unlock (&queue_lock);
+  pthread_mutex_lock (&queue_lock);
 
-	DEBUG ("rrdtool plugin: Put `%s' into the update queue", filename);
+  if (*tail == NULL)
+    *head = queue_entry;
+  else
+    (*tail)->next = queue_entry;
+  *tail = queue_entry;
 
-	return (0);
-} /* int rrd_queue_cache_entry */
+  pthread_cond_signal (&queue_cond);
+  pthread_mutex_unlock (&queue_lock);
+
+  return (0);
+} /* int rrd_queue_enqueue */
+
+static int rrd_queue_dequeue (const char *filename,
+    rrd_queue_t **head, rrd_queue_t **tail)
+{
+  rrd_queue_t *this;
+  rrd_queue_t *prev;
+
+  pthread_mutex_lock (&queue_lock);
+
+  prev = NULL;
+  this = *head;
+
+  while (this != NULL)
+  {
+    if (strcmp (this->filename, filename) == 0)
+      break;
+    
+    prev = this;
+    this = this->next;
+  }
+
+  if (this == NULL)
+  {
+    pthread_mutex_unlock (&queue_lock);
+    return (-1);
+  }
+
+  if (prev == NULL)
+    *head = this->next;
+  else
+    prev->next = this->next;
+
+  if (this->next == NULL)
+    *tail = prev;
+
+  pthread_mutex_unlock (&queue_lock);
+
+  sfree (this->filename);
+  sfree (this);
+
+  return (0);
+} /* int rrd_queue_dequeue */
 
 static void rrd_cache_flush (int timeout)
 {
@@ -726,13 +518,16 @@ static void rrd_cache_flush (int timeout)
 	iter = c_avl_get_iterator (cache);
 	while (c_avl_iterator_next (iter, (void *) &key, (void *) &rc) == 0)
 	{
-		if (rc->flags == FLAG_QUEUED)
+		if (rc->flags != FLAG_NONE)
 			continue;
 		else if ((now - rc->first_value) < timeout)
 			continue;
 		else if (rc->values_num > 0)
 		{
-			if (rrd_queue_cache_entry (key) == 0)
+			int status;
+
+			status = rrd_queue_enqueue (key, &queue_head,  &queue_tail);
+			if (status == 0)
 				rc->flags = FLAG_QUEUED;
 		}
 		else /* ancient and no values -> waste of memory */
@@ -778,6 +573,63 @@ static void rrd_cache_flush (int timeout)
 	cache_flush_last = now;
 } /* void rrd_cache_flush */
 
+static int rrd_cache_flush_identifier (int timeout, const char *identifier)
+{
+  rrd_cache_t *rc;
+  time_t now;
+  int status;
+  char key[2048];
+
+  if (identifier == NULL)
+  {
+    rrd_cache_flush (timeout);
+    return (0);
+  }
+
+  now = time (NULL);
+
+  if (datadir == NULL)
+    snprintf (key, sizeof (key), "%s.rrd",
+        identifier);
+  else
+    snprintf (key, sizeof (key), "%s/%s.rrd",
+        datadir, identifier);
+  key[sizeof (key) - 1] = 0;
+
+  status = c_avl_get (cache, key, (void *) &rc);
+  if (status != 0)
+  {
+    INFO ("rrdtool plugin: rrd_cache_flush_identifier: "
+        "c_avl_get (%s) failed. Does that file really exist?",
+        key);
+    return (status);
+  }
+
+  if (rc->flags == FLAG_FLUSHQ)
+  {
+    status = 0;
+  }
+  else if (rc->flags == FLAG_QUEUED)
+  {
+    rrd_queue_dequeue (key, &queue_head, &queue_tail);
+    status = rrd_queue_enqueue (key, &flushq_head, &flushq_tail);
+    if (status == 0)
+      rc->flags = FLAG_FLUSHQ;
+  }
+  else if ((now - rc->first_value) < timeout)
+  {
+    status = 0;
+  }
+  else if (rc->values_num > 0)
+  {
+    status = rrd_queue_enqueue (key, &flushq_head, &flushq_tail);
+    if (status == 0)
+      rc->flags = FLAG_FLUSHQ;
+  }
+
+  return (status);
+} /* int rrd_cache_flush_identifier */
+
 static int rrd_cache_insert (const char *filename,
 		const char *value, time_t value_time)
 {
@@ -787,6 +639,15 @@ static int rrd_cache_insert (const char *filename,
 
 	pthread_mutex_lock (&cache_lock);
 
+	/* This shouldn't happen, but it did happen at least once, so we'll be
+	 * careful. */
+	if (cache == NULL)
+	{
+		pthread_mutex_unlock (&cache_lock);
+		WARNING ("rrdtool plugin: cache == NULL.");
+		return (-1);
+	}
+
 	c_avl_get (cache, filename, (void *) &rc);
 
 	if (rc == NULL)
@@ -864,17 +725,20 @@ static int rrd_cache_insert (const char *filename,
 	}
 
 	DEBUG ("rrdtool plugin: rrd_cache_insert: file = %s; "
-			"values_num = %i; age = %u;",
+			"values_num = %i; age = %lu;",
 			filename, rc->values_num,
-			rc->last_value - rc->first_value);
+			(unsigned long)(rc->last_value - rc->first_value));
 
 	if ((rc->last_value - rc->first_value) >= cache_timeout)
 	{
 		/* XXX: If you need to lock both, cache_lock and queue_lock, at
 		 * the same time, ALWAYS lock `cache_lock' first! */
-		if (rc->flags != FLAG_QUEUED)
+		if (rc->flags == FLAG_NONE)
 		{
-			if (rrd_queue_cache_entry (filename) == 0)
+			int status;
+
+			status = rrd_queue_enqueue (filename, &queue_head, &queue_tail);
+			if (status == 0)
 				rc->flags = FLAG_QUEUED;
 		}
 		else
@@ -887,7 +751,6 @@ static int rrd_cache_insert (const char *filename,
 			((time (NULL) - cache_flush_last) > cache_flush_timeout))
 		rrd_cache_flush (cache_flush_timeout);
 
-
 	pthread_mutex_unlock (&cache_lock);
 
 	return (0);
@@ -913,6 +776,11 @@ static int rrd_write (const data_set_t *ds, const value_list_t *vl)
 	char         values[512];
 	int          status;
 
+	if (0 != strcmp (ds->type, vl->type)) {
+		ERROR ("rrdtool plugin: DS type does not match value list type");
+		return -1;
+	}
+
 	if (value_list_to_filename (filename, sizeof (filename), ds, vl) != 0)
 		return (-1);
 
@@ -923,7 +791,9 @@ static int rrd_write (const data_set_t *ds, const value_list_t *vl)
 	{
 		if (errno == ENOENT)
 		{
-			if (rrd_create_file (filename, ds, vl))
+			status = cu_rrd_create_file (filename,
+					ds, vl, &rrdcreate_config);
+			if (status != 0)
 				return (-1);
 		}
 		else
@@ -947,7 +817,7 @@ static int rrd_write (const data_set_t *ds, const value_list_t *vl)
 	return (status);
 } /* int rrd_write */
 
-static int rrd_flush (const int timeout)
+static int rrd_flush (int timeout, const char *identifier)
 {
 	pthread_mutex_lock (&cache_lock);
 
@@ -956,7 +826,8 @@ static int rrd_flush (const int timeout)
 		return (0);
 	}
 
-	rrd_cache_flush (timeout);
+	rrd_cache_flush_identifier (timeout, identifier);
+
 	pthread_mutex_unlock (&cache_lock);
 	return (0);
 } /* int rrd_flush */
@@ -970,6 +841,8 @@ static int rrd_config (const char *key, const char *value)
 		{
 			fprintf (stderr, "rrdtool: `CacheTimeout' must "
 					"be greater than 0.\n");
+			ERROR ("rrdtool: `CacheTimeout' must "
+					"be greater than 0.\n");
 			return (1);
 		}
 		cache_timeout = tmp;
@@ -981,6 +854,8 @@ static int rrd_config (const char *key, const char *value)
 		{
 			fprintf (stderr, "rrdtool: `CacheFlush' must "
 					"be greater than 0.\n");
+			ERROR ("rrdtool: `CacheFlush' must "
+					"be greater than 0.\n");
 			return (1);
 		}
 		cache_flush_timeout = tmp;
@@ -1007,15 +882,15 @@ static int rrd_config (const char *key, const char *value)
 	}
 	else if (strcasecmp ("StepSize", key) == 0)
 	{
-		stepsize = atoi (value);
-		if (stepsize < 0)
-			stepsize = 0;
+		int temp = atoi (value);
+		if (temp > 0)
+			rrdcreate_config.stepsize = temp;
 	}
 	else if (strcasecmp ("HeartBeat", key) == 0)
 	{
-		heartbeat = atoi (value);
-		if (heartbeat < 0)
-			heartbeat = 0;
+		int temp = atoi (value);
+		if (temp > 0)
+			rrdcreate_config.heartbeat = temp;
 	}
 	else if (strcasecmp ("RRARows", key) == 0)
 	{
@@ -1024,9 +899,11 @@ static int rrd_config (const char *key, const char *value)
 		{
 			fprintf (stderr, "rrdtool: `RRARows' must "
 					"be greater than 0.\n");
+			ERROR ("rrdtool: `RRARows' must "
+					"be greater than 0.\n");
 			return (1);
 		}
-		rrarows = tmp;
+		rrdcreate_config.rrarows = tmp;
 	}
 	else if (strcasecmp ("RRATimespan", key) == 0)
 	{
@@ -1045,23 +922,24 @@ static int rrd_config (const char *key, const char *value)
 		{
 			dummy = NULL;
 			
-			tmp_alloc = realloc (rra_timespans_custom,
-					sizeof (int) * (rra_timespans_custom_num + 1));
+			tmp_alloc = realloc (rrdcreate_config.timespans,
+					sizeof (int) * (rrdcreate_config.timespans_num + 1));
 			if (tmp_alloc == NULL)
 			{
 				fprintf (stderr, "rrdtool: realloc failed.\n");
+				ERROR ("rrdtool: realloc failed.\n");
 				free (value_copy);
 				return (1);
 			}
-			rra_timespans_custom = tmp_alloc;
-			rra_timespans_custom[rra_timespans_custom_num] = atoi (ptr);
-			if (rra_timespans_custom[rra_timespans_custom_num] != 0)
-				rra_timespans_custom_num++;
+			rrdcreate_config.timespans = tmp_alloc;
+			rrdcreate_config.timespans[rrdcreate_config.timespans_num] = atoi (ptr);
+			if (rrdcreate_config.timespans[rrdcreate_config.timespans_num] != 0)
+				rrdcreate_config.timespans_num++;
 		} /* while (strtok_r) */
 
-		qsort (/* base = */ rra_timespans_custom,
-				/* nmemb  = */ rra_timespans_custom_num,
-				/* size   = */ sizeof (rra_timespans_custom[0]),
+		qsort (/* base = */ rrdcreate_config.timespans,
+				/* nmemb  = */ rrdcreate_config.timespans_num,
+				/* size   = */ sizeof (rrdcreate_config.timespans[0]),
 				/* compar = */ rrd_compare_numeric);
 
 		free (value_copy);
@@ -1073,9 +951,30 @@ static int rrd_config (const char *key, const char *value)
 		{
 			fprintf (stderr, "rrdtool: `XFF' must "
 					"be in the range 0 to 1 (exclusive).");
+			ERROR ("rrdtool: `XFF' must "
+					"be in the range 0 to 1 (exclusive).");
 			return (1);
 		}
-		xff = tmp;
+		rrdcreate_config.xff = tmp;
+	}
+	else if (strcasecmp ("WritesPerSecond", key) == 0)
+	{
+		double wps = atof (value);
+
+		if (wps < 0.0)
+		{
+			fprintf (stderr, "rrdtool: `WritesPerSecond' must be "
+					"greater than or equal to zero.");
+			return (1);
+		}
+		else if (wps == 0.0)
+		{
+			write_rate = 0.0;
+		}
+		else
+		{
+			write_rate = 1.0 / wps;
+		}
 	}
 	else
 	{
@@ -1108,18 +1007,25 @@ static int rrd_shutdown (void)
 
 static int rrd_init (void)
 {
+	static int init_once = 0;
 	int status;
 
-	if (stepsize < 0)
-		stepsize = 0;
-	if (heartbeat <= 0)
-		heartbeat = 2 * stepsize;
+	if (init_once != 0)
+		return (0);
+	init_once = 1;
 
-	if ((heartbeat > 0) && (heartbeat < interval_g))
+	if (rrdcreate_config.stepsize < 0)
+		rrdcreate_config.stepsize = 0;
+	if (rrdcreate_config.heartbeat <= 0)
+		rrdcreate_config.heartbeat = 2 * rrdcreate_config.stepsize;
+
+	if ((rrdcreate_config.heartbeat > 0)
+			&& (rrdcreate_config.heartbeat < interval_g))
 		WARNING ("rrdtool plugin: Your `heartbeat' is "
 				"smaller than your `interval'. This will "
 				"likely cause problems.");
-	else if ((stepsize > 0) && (stepsize < interval_g))
+	else if ((rrdcreate_config.stepsize > 0)
+			&& (rrdcreate_config.stepsize < interval_g))
 		WARNING ("rrdtool plugin: Your `stepsize' is "
 				"smaller than your `interval'. This will "
 				"create needlessly big RRD-files.");
@@ -1155,7 +1061,10 @@ static int rrd_init (void)
 	DEBUG ("rrdtool plugin: rrd_init: datadir = %s; stepsize = %i;"
 			" heartbeat = %i; rrarows = %i; xff = %lf;",
 			(datadir == NULL) ? "(null)" : datadir,
-			stepsize, heartbeat, rrarows, xff);
+			rrdcreate_config.stepsize,
+			rrdcreate_config.heartbeat,
+			rrdcreate_config.rrarows,
+			rrdcreate_config.xff);
 
 	return (0);
 } /* int rrd_init */
